#include "operator.h"
#include <limits>

void Operand :: getInput ()
{ data = variable -> sourceOperator -> getOperand (variable -> sourceOperand) . data;
  data . transpose (transposition);
}


void Operand :: allocateData (vector <unsigned> & dimSizes)
{
  vector <int> sizes;
  if (in)  getInput ();
  else
  { unsigned s = 1;
    for (unsigned i = 0;  i < dimensions . size ();  i ++)
    { sizes . push_back (dimSizes [dimensions [i]]);
      s *= sizes . back ();
    }
    data = Slab (new Data [s], sizes);
  }
}


void Operator :: getDimSizes ()
{
  if (operands [0] . in)
    for (unsigned i = 0;  i < dimSizes . size ();  i ++)  dimSizes [i] = 0;
  for (unsigned i = 0;  i < operands . size ();  i ++)
  { Operand & o = operands [i];
    if (! o . in)  continue;
    for (unsigned j = 0;  j < o . dimensions . size ();  j ++)
    { unsigned thisDim = o . dimensions [j];
      unsigned size = 1;
      unsigned otherDim = o . variable -> sourceOperator
        -> operands [o . variable -> sourceOperand] . dimensions [o . transposition [j]];
      size = o . variable -> sourceOperator -> dimSizes [otherDim];

      if (dimSizes [thisDim] != 0 && dimSizes [thisDim] != size)
      { cerr << "dimsize mismatch (operand " << i << ", dim " << j << "): "
             << dimSizes [thisDim] << " != " << size << endl;
        exit (0);
      }

      dimSizes [thisDim] = size;
    }
  }
  calcDestSizes ();
}


double Operator :: calcCost (bool onlyIn)
{
  double cost = 0;
  for (unsigned o = 0;  o < operands . size ();  o ++)
  { if (! operands [o] . in || ! onlyIn)
      for (unsigned pi = 0;  pi < p;  pi ++)
        cost += calcCommCost (o, pi);
  }
  return cost * COMM_COST + calcCompCost ();
}


void Operator :: randomInit (unsigned pp)
{
  p = pp;
  unsigned prod = 1;
  getDimSizes ();
  setOutput ();
  for (unsigned i = 0;  i < splitDimensions;  i ++)
  { divisions [i] = (rand () % p) + 1;
    prod *= divisions [i];
  }
  for (unsigned i = 0;  i < splitDimensions;  i ++)
    divisions [i] = max (divisions [i] * p / prod, 1u);
  normalDivision ();
  responsibility . resize (p);
  randomPermutation (responsibility);
  bestDivisions = divisions;
  bestResponsibility = responsibility;
  calcDestSizes ();
}


bool Operator :: minimize ()
{
  bool betterFound = false;
  divisions = bestDivisions;
  responsibility = bestResponsibility;
  getDimSizes ();
  calcDestSizes ();
  double bestCost = calcCost ();
  for (beginDivision (p);  ! endDivision ();  nextDivision ())
  { calcDestSizes ();
    firstPermutation (responsibility);
    minimizeResponsibility ();
    double cost = calcCost ();
    if (cost < bestCost - 0.0001)
    { bestCost = cost;
      saveBest ();
      betterFound = true;
    }
  }
  restoreBest ();
  return betterFound;
}


bool Operator :: minimizeResponsibility ()
{
  bool changed = false;
  for (unsigned i = 0;  i < responsibility . size ();  i ++)
  { double bestCommCost = numeric_limits <double> :: max ();
    unsigned bestJ = 0;
    for (unsigned j = i;  j < responsibility . size ();  j ++)
    { double commCost = 0;
      swap (responsibility [i], responsibility [j]);
      for (unsigned o = 0;  o < operands . size ();  o ++)
        commCost += calcCommCost (o, i);
      swap (responsibility [i], responsibility [j]);
      if (commCost < bestCommCost)  {bestCommCost = commCost;  bestJ = j;}
    }
    if (i != bestJ)  changed = true;
    swap (responsibility [i], responsibility [bestJ]);
  }
  return changed;
}


void Operator :: setDivision (unsigned pp, string & scheme, unsigned & pos)
{
  p = pp;
  acceptToken (scheme, string (label) + ':' , pos);
  for (unsigned i = 0;  i < dimensions;  i ++)
  { if (i > 0)  acceptToken (scheme, ",", pos);
    readFromString (scheme, pos, divisions [i]);
  }
  responsibility . resize (p);
  for (unsigned i = 0;  i < p;  i ++)  responsibility [i] = i;
}


void Operator :: allocateData ()
{
#ifndef SLABBER_MPI

  for (unsigned o = 0;  o < operands . size ();  o ++)
    operands [o] . allocateData (dimSizes);

#else

  unsigned r = MPI :: COMM_WORLD . Get_rank ();
  vector <int> sizes;
  for (unsigned o = 0;  o < operands . size ();  o ++)
  { unsigned size = getRegion (o, r) . getSizes (sizes);
    operands [o] . data = Slab (new Data [size], sizes);
  }

#endif
}


void Operator :: getInput ()
{
#ifndef SLABBER_MPI

  for (unsigned o = 0;  o < operands . size ();  o ++)
    if (operands [o] . in)  operands [o] . getInput ();

#else

  unsigned r = MPI :: COMM_WORLD . Get_rank ();
  vector <MPI :: Request> requests;
  vector <MPI :: Datatype> datatypes;
  for (unsigned o = 0;  o < operands . size ();  o ++)
  { if (! operands [o] . in) continue;
    Variable & v = * operands [o] . variable;
    // Receive
    Slab slab = operands [o] . data;
    Region myReg = getRegion (o, r, true);
    slab . transpose (operands [o] . transposition);
    for (unsigned pi = 0;  pi < p;  pi ++)
    { if (pi == r)  continue;
      Region subReg
        =   myReg . intersect (v . sourceOperator -> getRegion (v . sourceOperand, pi))
          - myReg . getLow ();
      if (! subReg . empty ())
      { Slab subSlab = slab . subSlab (subReg . getLow (), subReg . getHigh ());
        datatypes . push_back (subSlab . createDatatype ());
        datatypes . back () . Commit ();
        requests . push_back
           (MPI :: COMM_WORLD . Irecv (subSlab . getData (), 1, datatypes . back (), pi, 0));
      }
    }
    // Send
    slab = v . sourceOperator -> operands [v . sourceOperand] . data;
    myReg = v . sourceOperator -> getRegion (v . sourceOperand, r);
    for (unsigned pi = 0;  pi < p;  pi ++)
    { if (pi == r)  continue;
      Region subReg = myReg . intersect (getRegion (o, pi, true)) - myReg . getLow ();
      if (! subReg . empty ())
      { Slab subSlab = slab . subSlab (subReg . getLow (), subReg . getHigh ());
        datatypes . push_back (subSlab . createDatatype ());
        datatypes . back () . Commit ();
        requests . push_back
           (MPI :: COMM_WORLD . Isend (subSlab . getData (), 1, datatypes . back (), pi, 0));
      }
    }
  }

  // local copy
  for (unsigned o = 0;  o < operands . size ();  o ++)
  { if (! operands [o] . in) continue;
    Variable & v = * operands [o] . variable;
    Slab toSlab = operands [o] . data;
    toSlab . transpose (operands [o] . transposition);
    Slab fromSlab = v . sourceOperator -> operands [v . sourceOperand] . data;
    Region toReg = getRegion (o, r, true);
    Region fromReg = v . sourceOperator -> getRegion (v . sourceOperand, r);
    Region toSubReg = toReg . intersect (fromReg) - toReg . getLow ();
    Region fromSubReg = fromReg . intersect (toReg) - fromReg . getLow ();
    if (! toSubReg . empty ())
    { Slab toSubSlab = toSlab . subSlab (toSubReg . getLow (), toSubReg . getHigh ());
      Slab fromSubSlab = fromSlab . subSlab (fromSubReg . getLow (), fromSubReg . getHigh ());
      toSubSlab . copyFrom (fromSubSlab);
    }
  }

  MPI :: Status status;
  for (unsigned i = 0;  i < requests . size ();  i ++)
    requests [i] . Wait (status);
  for (unsigned i = 0;  i < datatypes . size ();  i ++)
    datatypes [i] . Free ();

#endif
}



/*
void ReduceOperator :: reduce ()
{
  ThreadInfo * threadInfo
    = reinterpret_cast <ThreadInfo *> (pthread_getspecific (threadInfoKey));
  vector <unsigned> & reduceDivisionOrder = threadInfo -> reduceDivisionOrder;
  reduceDivisionOrder = divisionOrder;
  unsigned reduceDim = 0, destDim = 0;
  for (unsigned i = 0;  i < sourceDims; i ++)
  { if (reducedDims [i])  reduceDim = i;
    else  if (destDims [i] != 0)  destDim = i;
  }
  unsigned reduceDimOrder = reduceDivisionOrder [reduceDim];
  for (unsigned i = 0;  i < sourceDims;  i ++)
    if (reduceDivisionOrder [i] > reduceDimOrder)  reduceDivisionOrder [i] --;
  reduceDimOrder = reduceDivisionOrder [destDim];
  for (unsigned i = 0;  i < sourceDims;  i ++)
    if (reduceDivisionOrder [i] > reduceDimOrder)  reduceDivisionOrder [i] --;
  reduceDivisionOrder [reduceDim] = reduceDimOrder;

  exit (1);  // brauch ich reduce eigentlich??
}
*/

