#include <vector>
#include <iostream>
#ifdef SLABBER_MPI
#  include <mpi.h>
#endif

using namespace std;

typedef double Data;

template <typename T>
inline bool index2pos (unsigned index, vector <unsigned> & divisions, vector <T> & pos)
{
  pos . resize (divisions . size ());
  for (unsigned d = 0;  d < divisions . size ();  d ++)
  { pos [d] = index % divisions [d];
    index /= divisions [d];
  }
  return (index == 0);
}


template <typename T>
inline unsigned pos2index (vector <unsigned> & divisions, vector <T> & pos)
{
  unsigned index = 0;
  for (int i = divisions . size () - 1;  i >= 0;  i --)
    index = index * divisions [i] + pos [i];
  return index;
}


class Slab
{
public:
  Slab () : data (0)  {}

  Slab (Data * pdata, vector <int> & psize)
  : data (pdata), size (psize), stride (psize . size ())
  { int s = 1;
    //for (int j = size . size () - 1;  j >= 0;  j --)
    //  {stride [j] = s;  s *= size [j];}
    // row-major order for dimensions 0,1; and keep matrices together
    if (size . size () >= 2)  stride [1] = s;  s *= size [1];
    if (size . size () >= 1)  stride [0] = s;  s *= size [0];
    for (unsigned j = 2;  j < size . size ();  j ++)
      {stride [j] = s;  s *= size [j];}
  }

  Slab (Data * pdata, vector <int> & psize, vector <int> & pstride)
  : data (pdata), size (psize), stride (pstride)
  {}

  Data & operator[] (int i)  {return data [i * stride [0]];}
  Data & operator() (int i, int j)
    {return data [i * stride [0] + j * stride [1]];}
  Data & operator() (int i, int j, int k)
    {return data [i * stride [0] + j * stride [1] + k * stride [2]];}

  int getSize (unsigned i)  {return size [i];}

  Data * getData ()  {return data;}

  unsigned getDimensions ()  {return size . size ();}

  template <typename T1, typename T2>
  Slab slab (vector <T1> & sizes, vector <T2> pos)
  { vector <int> tsize;
    int p = 0;
    for (unsigned i = 0;  i < size . size ();  i ++)
    { tsize . push_back (max (0, min (int (sizes [i]), int (size [i] - pos [i] * sizes [i]))));
      p += pos [i] * sizes [i] * stride [i];
    }
    return Slab (data + p, tsize, stride);
  }

  Slab slab (unsigned index, vector <unsigned> & dimensions, vector <unsigned> & divisions, vector <int> * position = 0)
  {
    vector <int> sizes, pos, allpos;
    if (! index2pos (index, divisions, allpos))
    { for (unsigned d = 0;  d < divisions . size ();  d ++)  sizes . push_back (0);
      return Slab (0, sizes);
    }
    for (unsigned d = 0;  d < size . size ();  d ++)
    { unsigned div = divisions [dimensions [d]];
      sizes . push_back ((size [d] + div - 1) / div);
      pos . push_back (allpos [dimensions [d]]);
    }
    if (position)
    { (* position) . resize (pos . size ());
      for (unsigned d = 0;  d < pos . size ();  d ++)
        (* position) [d] = pos [d] * sizes [d];
    }
    return slab (sizes, pos);
  }

  template <typename T1, typename T2>
  Slab subSlab (vector <T1> const & low, vector <T2> const & high)
  {
    vector <int> sizes (size . size ());
    unsigned ind = 0;
    for (unsigned i = 0;  i < size . size ();  i ++)
    { sizes [i] = high [i] - low [i];
      ind += low [i] * stride [i];
    }
    return Slab (data + ind, sizes, stride);
  }

#ifdef SLABBER_MPI
  MPI :: Datatype createDatatype ()
  { MPI :: Datatype datatype = MPI :: DOUBLE;
    for (int i = size . size () - 1;  i >= 0;  i --)
      datatype = datatype . Create_hvector (size [i], 1, stride [i] * sizeof (double));
    return datatype;
  }
#endif

  void transpose (vector <unsigned> & transposition)
  { vector <int> osize = size;
    vector <int> ostride = stride;
    for (unsigned i = 0;  i < size . size ();  i ++)
    { size [i] = osize [transposition [i]];
      stride [i] = ostride [transposition [i]];
    }
  }

  void copyFrom (Slab & s)
  { if (size . size () == 1)
      for (int i = 0;  i < size [0];  i ++)  (* this) [i] = s [i];
    else if (size . size () == 2)
      for (int i = 0;  i < size [0];  i ++)
        for (int j = 0;  j < size [1];  j ++)
          (* this) (i, j) = s (i, j);
    else if (size . size () == 3)
      for (int i = 0;  i < size [0];  i ++)
        for (int j = 0;  j < size [1];  j ++)
          for (int k = 0;  k < size [2];  k ++)
          (* this) (i, j, k) = s (i, j, k);
    else  {cerr << "Slab-copy of dim > 3 or dim = 0 not implemented\n";  exit (1);}
  }

  int getStride (unsigned dim)  {return stride [dim];}

  void printFrame (ostream & o = cout)
  { //o << long long (data) << " (";
    o << '(';
    for (unsigned i = 0;  i < size . size ();  i ++)
    { if (i > 0)  o << ", ";
      o << size [i] << ':' << stride [i];
    }
    o << ')';
  }

  void print (ostream & o = cout)
  { if (size . size () == 2)
    { for (int i = 0;  i < size [0];  i ++)
      { for (int j = 0;  j < size [1];  j ++)  o << (* this) (i, j) << ' ';
        o << endl;
      }
    }
    if (size . size () == 3)
    { for (int i = 0;  i < size [0];  i ++)
      { for (int j = 0;  j < size [1];  j ++)
        { for (int k = 0;  k < size [1];  k ++)
            o << (* this) (i, j, k) << ' ';
          o << endl;
        }
        cout << "------\n";
      }
    }
  }

private:

  int calcSize ()
  { int s = 1;
    for (unsigned i = 0;  i < size . size ();  i ++)  s *= size [i];
    return s;
  }

  Data * data;
  vector <int> size;
  vector <int> stride;
};



#ifdef test_slab

int main ()
{
  int asize [] = {10, 20};  vector <int> size (asize, asize + 2);
  Slab s (0, size);
  s . printFrame ();
  unsigned adims [] = {0, 2};  vector <unsigned> dims (adims, adims + 2);
  unsigned adivs [] = {3, 3, 3};  vector <unsigned> divs (adivs, adivs + 3);
  unsigned adivo [] = {0, 1, 1};  vector <unsigned> divo (adivo, adivo + 3);
  for (unsigned i = 0;  i < 10;  i ++)
  { Slab t = s . slab (i, dims, divs, divo);
    t . printFrame ();
  }
}

#endif

